from itertools import combinations

import pharmpy.tools.modelfit as modelfit
from pharmpy.modeling import copy_model, fix_parameters_to
from pharmpy.modeling.block_rvs import create_joint_distribution
from pharmpy.workflows import Task, Workflow


def brute_force_no_of_etas(model):
    wf = Workflow()
    model_features = dict()

    param_names = [eta.parameter_names[0] for eta in model.random_variables.etas]
    param_combos = _get_combinations(param_names, include_single=True)

    for i, combo in enumerate(param_combos, 1):
        model_name = f'iiv_no_of_etas_candidate{i}'
        task_copy = Task('copy', copy, model_name)
        wf.add_task(task_copy)

        task_set_eta_to_zero = Task('set_eta_to_zero', set_eta_to_zero, combo)
        wf.add_task(task_set_eta_to_zero, predecessors=task_copy)

        model_features[model_name] = combo

    wf_fit = modelfit.create_workflow(n=i)
    wf.insert_workflow(wf_fit)
    return wf, model_features


def set_eta_to_zero(params, model):
    fix_parameters_to(model, params, 0)
    return model


def brute_force_block_structure(model):
    wf = Workflow()
    model_features = dict()

    eta_combos_single_blocks, eta_combos_multi_blocks = _get_possible_iiv_blocks(model)
    eta_combos_all = eta_combos_single_blocks + eta_combos_multi_blocks

    for i, combo in enumerate(eta_combos_all, 1):
        model_name = f'iiv_block_structure_candidate{i}'
        task_copy = Task('copy', copy, model_name)
        wf.add_task(task_copy)

        task_joint_dist = Task('create_joint_dist', create_joint_dist, combo)
        wf.add_task(task_joint_dist, predecessors=task_copy)

        model_features[model_name] = combo

    wf_fit = modelfit.create_workflow(n=len(eta_combos_all))
    wf.insert_workflow(wf_fit)
    return wf, model_features


def _get_combinations(names, include_single=False):
    combos = []
    if include_single:
        start = 1
    else:
        start = 2
    for i in range(start, len(names) + 1):
        combos += [list(combo) for combo in combinations(names, i)]
    return combos


def _get_possible_iiv_blocks(model):
    eta_names = model.random_variables.etas.names
    eta_combos_single_blocks = _get_combinations(eta_names)
    if len(eta_names) < 4:
        return eta_combos_single_blocks, []

    no_of_blocks_max = int(len(eta_names) / 2)
    eta_combos_multi_blocks = []

    for i in range(2, no_of_blocks_max + 1):
        etas = []
        for combo in combinations(eta_combos_single_blocks, i):
            combo = list(combo)
            if len(flatten(combo)) == len(eta_names) and len(set(flatten(combo))) == len(eta_names):
                etas.append(combo)
        eta_combos_multi_blocks += etas

    return eta_combos_single_blocks, eta_combos_multi_blocks


def flatten(list_to_flatten):
    return [item for sublist in list_to_flatten for item in sublist]


def create_joint_dist(list_of_etas, model):
    if isinstance(list_of_etas[0], list):
        for eta_block in list_of_etas:
            create_joint_distribution(model, eta_block)
    else:
        create_joint_distribution(model, list_of_etas)
    return model


def copy(name, model):
    model_copy = copy_model(model, name)
    return model_copy



import numpy as np

# All functions used for comparing a set of candidate models and
# ranking them take the same base arguments
# base - the base model
# candidates - a list of candidate models
# returns a list which is a subset of the candidates with the best ranked at the beginning
# models can be removed from candidates if they don't pass some criteria


def ofv(base, candidates, cutoff=3.84):
    try:
        base_ofv = base.modelfit_results.ofv
    except AttributeError:
        base_ofv = np.nan
    return rank_models('ofv', base_ofv, candidates, cutoff)


def aic(base, candidates, cutoff=3.84):
    try:
        base_aic = base.modelfit_results.aic
    except AttributeError:
        base_aic = np.nan
    return rank_models('aic', base_aic, candidates, cutoff)


def bic(base, candidates, cutoff=3.84):
    try:
        base_bic = base.modelfit_results.bic
    except AttributeError:
        base_bic = np.nan
    return rank_models('bic', base_bic, candidates, cutoff)


def rank_models(rank_type, base_value, candidates, cutoff):
    filtered = [
        model
        for model in candidates
        if model.modelfit_results is not None
        and base_value - getattr(model.modelfit_results, rank_type) >= cutoff
    ]
    srtd = sorted(filtered, key=lambda model: getattr(model.modelfit_results, rank_type))
    return srtd



import shutil

import numpy as np
import pytest

from pharmpy import Model
from pharmpy.modeling import add_iiv, fit, run_tool
from pharmpy.utils import TemporaryDirectoryChanger


@pytest.mark.parametrize(
    'mfl, no_of_models, best_model_name',
    [
        ('ABSORPTION(ZO)\nPERIPHERALS(1)', 4, 'modelsearch_candidate2'),
        ('ABSORPTION(ZO)\nTRANSITS(1)', 2, 'mox2'),
        ('ABSORPTION([ZO,SEQ-ZO-FO])\nPERIPHERALS(1)', 7, 'modelsearch_candidate3'),
        ('LAGTIME()\nTRANSITS(1)', 2, 'mox2'),
    ],
)
def test_exhaustive_stepwise(tmp_path, testdata, mfl, no_of_models, best_model_name):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mx19B.csv', tmp_path)
        model_start = Model('mox2.mod')
        model_start.dataset_path = tmp_path / 'mx19B.csv'
        res = run_tool('modelsearch', 'exhaustive_stepwise', mfl, model=model_start)

        assert len(res.summary) == no_of_models
        assert len(res.models) == no_of_models
        assert all(
            model.modelfit_results and not np.isnan(model.modelfit_results.ofv)
            for model in res.models
        )
        assert res.best_model.name == best_model_name
        rundir = tmp_path / 'modelsearch_dir1'
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == no_of_models + 1

        assert res.to_json()


@pytest.mark.parametrize(
    'mfl, as_fullblock, no_of_models, best_model_name, no_of_added_etas',
    [
        ('ABSORPTION(ZO)\nPERIPHERALS(1)', False, 4, 'modelsearch_candidate2', 2),
        ('ABSORPTION(ZO)\nPERIPHERALS([1, 2])', False, 7, 'modelsearch_candidate2', 4),
        ('LAGTIME()\nTRANSITS(1)', False, 2, 'mox2', 1),
        ('ABSORPTION(ZO)\nPERIPHERALS(1)', True, 4, 'modelsearch_candidate2', 2),
    ],
)
def test_exhaustive_stepwise_add_etas(
    tmp_path, testdata, mfl, as_fullblock, no_of_models, best_model_name, no_of_added_etas
):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mx19B.csv', tmp_path)
        model_start = Model('mox2.mod')
        model_start.dataset_path = tmp_path / 'mx19B.csv'
        res = run_tool('modelsearch', 'exhaustive_stepwise', mfl, add_etas=True, model=model_start)

        assert len(res.summary) == no_of_models
        assert len(res.models) == no_of_models
        assert res.best_model.name == best_model_name
        model_last = res.models[no_of_models - 1]
        assert (
            len(model_last.random_variables.etas) - len(model_start.random_variables.etas)
            == no_of_added_etas
        )
        rundir = tmp_path / 'modelsearch_dir1'
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == no_of_models + 1


def test_exhaustive_stepwise_already_fit(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mx19B.csv', tmp_path)
        model_start = Model('mox2.mod')
        model_start.dataset_path = tmp_path / 'mx19B.csv'

        fit(model_start)

        mfl = 'ABSORPTION(ZO)\nPERIPHERALS(1)'
        res = run_tool('modelsearch', 'exhaustive_stepwise', mfl, model=model_start)

        assert len(res.summary) == 4
        assert len(res.models) == 4
        assert all(
            model.modelfit_results and not np.isnan(model.modelfit_results.ofv)
            for model in res.models
        )
        rundir = tmp_path / 'modelsearch_dir1'
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == 4


def test_exhaustive_stepwise_start_model_fail(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mx19B.csv', tmp_path)
        model_start = Model('mox2.mod')
        model_start.dataset_path = tmp_path / 'mx19B.csv'

        add_iiv(model_start, 'V', 'incorrect_syntax')

        mfl = 'ABSORPTION(ZO)\nPERIPHERALS(1)'
        res = run_tool('modelsearch', 'exhaustive_stepwise', mfl, model=model_start)

        assert len(res.summary) == 4
        assert res.summary['dofv'].isnull().values.all()
        assert len(res.models) == 4
        assert all(model.modelfit_results is None for model in res.models)
        rundir = tmp_path / 'modelsearch_dir1'
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == 5



import shutil

from pharmpy import Model
from pharmpy.modeling import run_iiv, run_tool
from pharmpy.utils import TemporaryDirectoryChanger


def test_amd(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        for path in (testdata / 'nonmem').glob('pheno_real.*'):
            shutil.copy2(path, tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'sdtab1', tmp_path)

        model = Model('pheno_real.mod')
        model.dataset_path = tmp_path / 'pheno.dta'

        # FIXME: remove after updating results
        model.modelfit_results.estimation_step

        res = run_tool('amd', model)
        assert res


def test_iiv(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mx19B.csv', tmp_path)
        model_start = Model('mox2.mod')
        model_start.dataset_path = tmp_path / 'mx19B.csv'

        res = run_iiv(model_start)

        assert len(res.summary) == 11
        assert len(res.models) == 11
        rundir1 = tmp_path / 'iiv_dir1'
        assert rundir1.is_dir()
        assert len(list((rundir1 / 'models').iterdir())) == 8
        rundir2 = tmp_path / 'iiv_dir2'
        assert rundir2.is_dir()
        assert len(list((rundir2 / 'models').iterdir())) == 4



import pandas as pd

from pharmpy.modeling import copy_model, remove_iiv
from pharmpy.results import Results
from pharmpy.tools.iiv.tool import IIVResults
from pharmpy.workflows import default_tool_database

from .run import run_tool


class AMDResults(Results):
    def __init__(self, final_model=None):
        self.final_model = final_model


def run_amd(model):
    db = default_tool_database(toolname='amd')
    run_tool('modelfit', model, path=db.path / 'modelfit')

    mfl = 'LAGTIME()\nPERIPHERALS(1)'
    res_modelsearch = run_tool(
        'modelsearch', 'exhaustive_stepwise', mfl=mfl, rankfunc='ofv', cutoff=3.84, model=model
    )
    selected_model = res_modelsearch.best_model

    res_iiv = run_tool('iiv', 'brute_force', rankfunc='ofv', cutoff=3.84, model=selected_model)
    selected_iiv_model = res_iiv.best_model

    res_resmod = run_tool('resmod', selected_iiv_model)
    final_model = res_resmod.best_model

    res = AMDResults(final_model=final_model)

    return res


def run_iiv(model):
    res_no_of_etas = run_tool('iiv', 'brute_force_no_of_etas', model=model)

    if res_no_of_etas.best_model != model:
        best_model = res_no_of_etas.best_model
        best_model_eta_removed = copy_model(res_no_of_etas.best_model, 'iiv_no_of_etas_best_model')
        features = res_no_of_etas.summary.loc[best_model.name]['features']
        remove_iiv(best_model_eta_removed, features)
    else:
        best_model = model

    res_block_structure = run_tool('iiv', 'brute_force_block_structure', model=best_model)

    best_model = res_block_structure.best_model
    summary = pd.concat([res_no_of_etas.summary, res_block_structure.summary])
    res = IIVResults(
        summary=summary,
        best_model=best_model,
        models=res_no_of_etas.models + res_block_structure.models,
        start_model=model,
    )

    return res



import shutil

import pytest

from pharmpy import Model
from pharmpy.modeling import add_iiv, add_peripheral_compartment, run_tool
from pharmpy.utils import TemporaryDirectoryChanger


@pytest.mark.parametrize(
    'list_of_parameters, no_of_models',
    [([], 4), (['QP1'], 14)],
)
def test_iiv_block_structure(tmp_path, testdata, list_of_parameters, no_of_models):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mx19B.csv', tmp_path)
        model_start = Model('mox2.mod')
        model_start.dataset_path = tmp_path / 'mx19B.csv'

        add_peripheral_compartment(model_start)
        add_iiv(model_start, list_of_parameters, 'add')

        res = run_tool('iiv', 'brute_force_block_structure', model=model_start)

        assert len(res.summary) == no_of_models
        assert len(res.models) == no_of_models
        assert all(int(model.modelfit_results.ofv) in range(-2200, -2000) for model in res.models)
        rundir = tmp_path / 'iiv_dir1'
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == no_of_models + 1


def test_iiv_no_of_etas(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mx19B.csv', tmp_path)
        model_start = Model('mox2.mod')
        model_start.dataset_path = tmp_path / 'mx19B.csv'

        res = run_tool('iiv', 'brute_force_no_of_etas', model=model_start)

        assert len(res.summary) == 7
        assert len(res.models) == 7
        assert all(int(model.modelfit_results.ofv) in range(-1900, -1400) for model in res.models)
        rundir = tmp_path / 'iiv_dir1'
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == 8
